import base64
import configparser
import gc
import inspect
import json
import logging
import os
import queue
import shutil
import socket
import threading
import time
from abc import ABC, abstractmethod

import librosa
import numpy as np

from openmmla_audio.base._enum import BLUE, ENDC, GREEN
from openmmla_audio.utils.audio_recognizer import AudioRecognizer
from openmmla_audio.utils.audio_recorder import AudioRecorder
from openmmla_audio.utils.audio_utils import read_byte_from_wav, write_frames_to_file
from openmmla_audio.utils.auga import normalize_rms
from openmmla_audio.utils.error_utils import RecordingError, TranscribingError
from openmmla_audio.utils.influx_client import InfluxDBClientWrapper
from openmmla_audio.utils.input_utils import get_function_base, get_id
from openmmla_audio.utils.logger_utils import get_logger
from openmmla_audio.utils.mqtt_client import MQTTClientWrapper
from openmmla_audio.utils.os_utils import clear_directory
from openmmla_audio.utils.redis_client import RedisClientWrapper
from openmmla_audio.utils.request_utils import request_speech_transcription, request_speech_separation
from openmmla_audio.utils.thread_utils import RaisingThread

try:
    from openmmla_audio.utils.transcriber import Transcriber
except ImportError:
    Transcriber = None

try:
    import torch
except ImportError:
    torch = None

try:
    from modelscope.pipelines import pipeline
    from modelscope.utils.constant import Tasks
except ImportError:
    pipeline, Tasks = None, None


class AudioBase(ABC):
    logger = get_logger(f'audio-base')

    def __init__(self, project_dir: str = None, config_path: str = None, mode: str = 'full', local: bool = False,
                 vad: bool = True, nr: bool = True, tr: bool = True, sp: bool = False, store: bool = True):
        """Initialize the Audio Base.

        Args:
            project_dir: root directory of the project, default to the directory of the caller module when not specified
            config_path: path to the configuration file, either absolute or relative to the root directory, default to
                            'conf/audio_base.ini' when not specified
            mode: operating mode, either 'record', 'recognize', or 'full', default to 'full'
            local: whether to run the audio base locally, default to False
            vad: whether to use the VAD, default to True
            nr: whether to use the denoiser to enhance speech, default to True
            tr: whether to transcribe speech to text, default to True
            sp: whether to do speech separation for overlapped segment, default to False
            store: whether to store audio files, default to True
        """
        # Set the project root directory
        if project_dir is None:
            caller_frame = inspect.stack()[1]
            caller_module = inspect.getmodule(caller_frame[0])
            if caller_module is None:
                self.project_dir = os.getcwd()
            else:
                self.project_dir = os.path.dirname(os.path.abspath(caller_module.__file__))
        else:
            self.project_dir = project_dir

        # Determine the configuration path
        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(self.project_dir, config_path)
        else:
            self.config_path = os.path.join(self.project_dir, 'conf/audio_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        self.mode = mode
        self.local = local
        self.vad = vad
        self.nr = nr
        self.tr = tr
        self.sp = sp
        self.store = store

        # Set Directories
        self.audio_realtime_dir = os.path.join(self.project_dir, 'audio', 'real-time')
        self.audio_temp_dir = os.path.join(self.project_dir, 'audio', 'temp')
        self.audio_db_dir = os.path.join(self.project_dir, 'audio_db')
        os.makedirs(self.audio_realtime_dir, exist_ok=True)
        os.makedirs(self.audio_temp_dir, exist_ok=True)
        os.makedirs(self.audio_db_dir, exist_ok=True)

        # Runtime variables
        self.bucket_name = None
        self.last_speaker = None
        self.audio_dir = None
        self.audio_queue = None
        self.transcription_queue = None
        self.speaker_frames_dict = None
        self.threads = []
        self.stop_event = threading.Event()

        # Attributes to be initialized in the subclass
        self.register_duration = None  # segment length of speaker registration
        self.recognize_duration = None  # segment length of speaker recognition
        self.rms_threshold = None  # volume energy level RMS threshold
        self.rms_peak_threshold = None  # volume energy level RMS peak threshold
        self.threshold = None  # similarity threshold for speaker recognition
        self.keep_threshold = None  # similarity threshold for keep speaker when doing half-scaled recognition
        self.audio_server_host = None  # audio server host IP

        self.config = None  # config object
        self.influx_client = None  # InfluxDB client object
        self.redis_client = None  # Redis client object
        self.mqtt_client = None  # MQTT client object
        self.audio_recognizer = None  # audio recognizer object
        self.audio_recorder = None  # audio recorder object
        self.speech_transcriber = None  # speech transcriber object
        self.speech_separator = None  # speech separator object

        self.id = None  # audio base id
        self.audio_db = None  # audio database directory

        self._load_config()
        self._config_from_user()
        self._config_objects()

    @property
    @abstractmethod
    def base_type(self):
        """Return the type of the audio base."""
        pass

    def _load_config(self):
        """Configure the audio base based on the arguments passed from the subclass."""
        self.config = configparser.ConfigParser()
        self.config.read(self.config_path)
        self.register_duration = int(self.config[self.base_type]['register_duration'])
        self.rms_threshold = int(self.config[self.base_type]['rms_threshold'])
        self.rms_peak_threshold = int(self.config[self.base_type]['rms_peak_threshold'])
        self.threshold = float(self.config[self.base_type]['recognize_sp_threshold']) if self.sp else float(
            self.config[self.base_type]['recognize_threshold'])
        self.keep_threshold = float(self.config[self.base_type]['keep_sp_threshold']) if self.sp else float(
            self.config[self.base_type]['keep_threshold'])
        self.recognize_duration = int(self.config[self.base_type]['recognize_sp_duration']) if self.sp else int(
            self.config[self.base_type]['recognize_duration'])
        self.audio_server_host = socket.gethostbyname(self.config['Server']['audio_server_host'])

    def _config_from_user(self):
        """Configure the audio base based on the user input."""
        self.id = get_id()
        self.audio_db = os.path.join(self.audio_db_dir, f'{self.base_type}_{self.id}')
        os.makedirs(self.audio_db, exist_ok=True)

    def _config_objects(self):
        """Configure the audio base based on the configuration file."""
        self.influx_client = InfluxDBClientWrapper(self.config_path)
        self.redis_client = RedisClientWrapper(self.config_path)
        self.mqtt_client = MQTTClientWrapper(self.config_path)
        self.warm_up_resampler()

        if self.local:
            self.audio_recorder = AudioRecorder(config_path=self.config_path, vad_enable=self.vad, nr_enable=self.nr,
                                                nr_local=True, vad_local=True)
            self.audio_recognizer = AudioRecognizer(config_path=self.config_path, audio_db=self.audio_db, local=True,
                                                    model_path=self.config['Local']['sr_model'])
            if self.tr:
                if Transcriber is None:
                    raise ImportError("Transcriber module is not available, please install the required dependencies.")
                self.speech_transcriber = Transcriber(self.config['Local']['tr_model'],
                                                      self.config['Local']['language'])

            if self.sp:
                if pipeline is None or Tasks is None:
                    raise ImportError("Modelscope module is not available, please install the required dependencies.")
                logging.getLogger('modelscope').setLevel(logging.WARNING)
                device = 'gpu' if torch.cuda.is_available() else 'cpu'
                try:
                    self.speech_separator = pipeline(Tasks.speech_separation, device=device,
                                                     model=self.config['Local']['sp_model'])
                except ValueError:
                    self.separator = pipeline(Tasks.speech_separation, device=device,
                                              model=self.config['Local']['sp_model_local'])
        else:
            self.audio_recorder = AudioRecorder(config_path=self.config_path, vad_enable=self.vad, nr_enable=self.nr)
            self.audio_recognizer = AudioRecognizer(config_path=self.config_path, audio_db=self.audio_db)

    def run(self):
        """Interface for running the Audio Base."""
        func_map = {1: self._register_profile, 2: self._recognize_voice, 3: self._reset, 4: self._switch_mode}
        while True:
            try:
                print(f"\033]0;Audio Base {self.base_type} {self.id} \007")
                select_fun = get_function_base()
                if select_fun == 0:
                    print("------------------------------------------------")
                    clear_directory(self.audio_temp_dir)
                    self.logger.info("Exiting the program...")
                    break
                func_map.get(select_fun, lambda: self.logger.warning("Invalid option"))()
            except (Exception, KeyboardInterrupt) as e:
                self._clean_up()
                self.logger.warning(
                    f"During running the audio base, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}, Come back to the main menu.",
                    exc_info=True)

    @abstractmethod
    def _register_profile(self):
        """Register participant's voice to audio database."""
        pass

    @abstractmethod
    def _recognize_voice(self):
        """Start the real-time voice recognition, including audio recording, speaker recognition, speech transcription,
        and listening on stop signal."""
        pass

    @abstractmethod
    def _reset(self):
        """Set the new port for the audio base and reinitialize audio base."""
        pass

    @abstractmethod
    def _switch_mode(self):
        """Switch the mode of the audio base."""
        pass

    def _create_thread(self, target, *args):
        """Create a new thread and add it to the thread list.

        Args:
            target: the target function to run in the thread
            *args: the arguments to pass to Thread
        """
        t = RaisingThread(target=target, args=args)
        self.threads.append(t)

    def _start_threads(self):
        """Start all threads."""
        self.stop_event.clear()
        for t in self.threads:
            t.start()

    def _join_threads(self):
        """Wait for all threads to finish."""
        for t in self.threads:
            t.join()

    def _stop_threads(self):
        """Stop all threads and free memory."""
        self.stop_event.set()
        for t in self.threads:
            if threading.current_thread() != t:
                try:
                    t.join(timeout=5)
                except Exception as e:
                    self.logger.warning(f"During the thread stopping, catch: {e}")
        self.threads.clear()

    def _clean_up(self):
        """Free memory by resetting the runtime variables."""
        self.bucket_name = None
        self.last_speaker = None
        self.audio_dir = None
        self.audio_queue = None
        self.transcription_queue = None
        self.speaker_frames_dict = None
        gc.collect()

    def _listen_for_start_signal(self):
        """Listen on the redis bucket control channel for the START signal."""
        p = self.redis_client.subscribe(f'{self.bucket_name}/control')  # Note: pubsub is not thread-safe
        self.logger.info("Wait for START signal...")
        while True:
            message = p.get_message()
            if message and message['data'] == b'START':
                self.logger.info("Received START signal, start recognizing...")
                break

    def _listen_for_stop_signal(self):
        """Listen on the redis bucket control channel for the STOP signal."""
        p = self.redis_client.subscribe(f'{self.bucket_name}/control')
        self.logger.info("Listening for STOP signal...")
        while not self.stop_event.is_set():
            message = p.get_message()
            if message and message['data'] == b'STOP':
                self.logger.info("Received STOP signal, stop recognizing...")
                self._stop_threads()

    def _load_and_queue_recorded_files(self):
        """Load pre-recorded audio files and queue them for recognition when in recognize mode."""
        while not self.stop_event.is_set():
            try:
                print(f"{GREEN}[Pre-recorded Audio]{ENDC}Loading pre-recorded audio files...")
                audio_files = [os.path.join(self.audio_dir, 'records', f) for f in
                               os.listdir(os.path.join(self.audio_dir, 'records'))
                               if f.endswith('.wav')]
                audio_files_sorted = sorted(audio_files, key=lambda x: os.path.getmtime(x))

                for file_path in audio_files_sorted:
                    frames = read_byte_from_wav(file_path)
                    temp_file_path = os.path.join(self.audio_dir, 'temp', os.path.basename(file_path))
                    self.audio_queue.put((temp_file_path, frames))
                print(f"{GREEN}[Pre-recorded Audio]{ENDC}Pre-recorded audio files loaded and queued.")
                return
            except Exception as e:
                raise RecordingError(
                    f'RecordingError occurred when loading pre-recorded audio files and queue: {e}') from e

    def _continuous_transcribing(self):
        """Continuously transcribe audio from the transcription queue."""
        while not self.stop_event.is_set():
            try:
                frames, speaker, chunk_start_time, chunk_end_time = self.transcription_queue.get(timeout=2)
                text = self._transcribe(frames)
                self._upload_transcription_data(speaker, text, chunk_start_time, chunk_end_time)
            except queue.Empty:
                continue
            except Exception as e:
                raise TranscribingError(f'TranscribingError occurred when transcribing: {e}') from e

    def _update_chunk_list(self, left_speaker, right_speaker, last_speaker, current_speaker, chunk_frames, frames,
                           record_start_time):
        """Update the chunk list based on the half-scaled recognition results at the speaker turn border.

        Args:
            left_speaker: recognized speaker of the left half-segment
            right_speaker: recognized speaker of the right half-segment
            last_speaker: recognized speaker of the last segment
            current_speaker: recognized speaker of the current segment
            chunk_frames: audio frames of the current chunk belonging to the last speaker
            frames: audio frames of the current segment
            record_start_time: record start time of the current segment
        """
        num_frames = int(len(frames) / 2)

        if left_speaker == right_speaker and left_speaker == last_speaker:  # Case AAAB: Extend the left chunk
            chunk_frames = chunk_frames + frames[:num_frames]
            frames = frames[num_frames:]
            record_start_time = str(float(record_start_time) + self.recognize_duration / 2)
            chunk_end_time = record_start_time
        elif left_speaker == right_speaker and right_speaker == current_speaker:  # Case ABBB: Extend the right chunk
            frames = chunk_frames[-num_frames:] + frames
            chunk_frames = chunk_frames[:-num_frames]
            record_start_time = str(float(record_start_time) - self.recognize_duration / 2)
            chunk_end_time = record_start_time
        else:  # Other cases, AABB, A_BB, AA_B, A__B (excluding ABAB, AB_B, A_AB)
            chunk_end_time = record_start_time
            if left_speaker != last_speaker:
                chunk_frames = chunk_frames[:-num_frames]
                chunk_end_time = str(float(record_start_time) - self.recognize_duration / 2)
            if right_speaker != current_speaker:
                frames = frames[num_frames:]
                record_start_time = str(float(record_start_time) + self.recognize_duration / 2)

        return chunk_frames, frames, record_start_time, chunk_end_time

    def _add_to_transcription_queue(self, frames, speaker, chunk_start_time, chunk_end_time):
        """Add transcription data to the transcription queue.

        Args:
            frames: audio frames to be transcribed
            speaker: recognized speaker of the audio frames
            chunk_start_time: start time of the audio frames
            chunk_end_time: end time of the audio frames
        """
        if self.tr:
            self.transcription_queue.put((frames, speaker, chunk_start_time, chunk_end_time))

    def _transcribe(self, frames):
        """Transcribe audio frames to text.

        Args:
            frames: audio frames to be transcribed

        Returns:
            transcribed text
        """
        if self.local:
            fr = 8000 if self.sp else 16000
            audio_file_path = os.path.join(self.audio_temp_dir, f'transcribe_audio_{self.base_type}_{self.id}.wav')
            write_frames_to_file(output_path=audio_file_path, frames=frames, channels=1, sampwidth=2, framerate=fr)
            if fr == 16000:
                self.audio_recorder.apply_nr(audio_file_path)
            normalize_rms(infile=audio_file_path, rms_level=-20)
            text = self.speech_transcriber.transcribe(audio_file_path)
        else:
            text = request_speech_transcription(frames, f'{self.base_type.lower()}_{self.id}', self.sp,
                                                self.audio_server_host)
        return text

    def _separate_speech(self, segment_audio_path):
        """Separate speech from the overlapped segment audio.

        Args:
            segment_audio_path:

        Returns:
            separated speech signals
        """
        if self.local:
            separated_result = self.speech_separator(segment_audio_path)
            result = separated_result['output_pcm_list']
        else:
            separated_result = request_speech_separation(segment_audio_path, f'{self.base_type.lower()}_{self.id}',
                                                         self.audio_server_host)
            result = [base64.b64decode(encoded_bytes_stream) for encoded_bytes_stream in separated_result]
        return result

    def _upload_transcription_data(self, speaker, text, chunk_start_time, chunk_end_time):
        """Upload transcription data to InfluxDB.

        Args:
            speaker: recognized speaker name of the transcribed chunk
            text: transcribed text of the chunk
            chunk_start_time: start time of the chunk
            chunk_end_time:  end time of the chunk
        """
        transcription_data = {
            "measurement": "speaker transcription",
            "fields": {
                "chunk_start_time": float(chunk_start_time),
                "chunk_end_time": float(chunk_end_time),
                "text": text,
                "speaker": speaker,
            },
        }
        print(f"{GREEN}[Speaker Transcription]{ENDC}{transcription_data['fields']['chunk_start_time']}: "
              f"{GREEN}{speaker} : {text}{ENDC}")
        self.influx_client.write(self.bucket_name, record=transcription_data)

    def _log_and_publish_results(self, record_start_time, recognize_start_time, speakers, similarities, durations):
        """Log and publish base speaker recognition results in Json string to Redis on bucket channel.

        Args:
            record_start_time: record start time of the segment
            recognize_start_time: start time of the recognition process
            speakers: recognized speakers
            similarities: similarity values
            durations: audio durations
        """
        base_recognition_result = {
            'base_id': f'{self.base_type.lower()}_{self.id}',
            'record_start_time': record_start_time,
            'speakers': json.dumps(speakers),
            'similarities': json.dumps(similarities),
            'durations': json.dumps(durations)
        }
        print(f"{BLUE}[Speaker Recognition]{ENDC}{base_recognition_result['record_start_time']}: "
              f"{BLUE}{base_recognition_result['speakers']}{ENDC}, similarity: {base_recognition_result['similarities']},"
              f"processed time: {time.time() - recognize_start_time} seconds")
        result_str = json.dumps(base_recognition_result)
        self.mqtt_client.publish(f'{self.bucket_name}/audio', result_str)

    def _prepare_directories(self):
        """Prepare and manage directories based on the operation mode."""
        sub_dirs = ['segments', 'chunks', 'separations', 'temp', 'records']
        for subdir in sub_dirs:
            directory_path = os.path.join(self.audio_dir, subdir)
            os.makedirs(directory_path, exist_ok=True)

        if self.mode == 'recognize':
            for subdir in ['segments', 'chunks', 'separations']:
                clear_directory(os.path.join(self.audio_dir, subdir))

        if self.mode == 'record':
            clear_directory(os.path.join(self.audio_dir, 'records'))

        clear_directory(os.path.join(self.audio_dir, 'temp'))

    def _store_audio(self, source_path, dest_path):
        """Store or remove audio files based on the store flag.

        Args:
            source_path: original audio file path
            dest_path: destination audio file path
        """
        if self.store:
            shutil.move(source_path, dest_path)
        else:
            os.remove(source_path)

    def _finalize_sp_results(self, speakers, similarities, durations, signals):
        """Post-process the recognition results of separated signals to handle edge cases.

        Args:
            speakers: list of recognized speakers
            similarities: list of similarity values
            durations: list of audio durations
            signals: list of separated signals

        Returns:
            post-processed speakers, similarities, durations, signals
        """
        # If there's only one speaker, keep it as is
        if len(speakers) == 1:
            return speakers, similarities, durations, signals

        # If both speakers are the same, keep the one with the highest similarity
        if speakers[0] == speakers[1]:
            max_similarity_index = 0 if similarities[0] > similarities[1] else 1
            return [speakers[max_similarity_index]], [similarities[max_similarity_index]], [
                durations[max_similarity_index]], [signals[max_similarity_index]]

        # If both are real speakers, keep them as is
        if all(speaker not in ['silent', 'unknown'] for speaker in speakers):
            return speakers, similarities, durations, signals

        # If there's at least one real speaker, keep only the real ones
        real_speakers, real_similarities, real_durations, real_texts = self.filter_real_speakers(speakers, similarities,
                                                                                                 durations, signals)
        if real_speakers:
            return real_speakers, real_similarities, real_durations, real_texts

        # Special cases for 'silent' and 'unknown'
        if 'unknown' in speakers and 'silent' in speakers:
            return (['unknown'], [similarities[speakers.index('unknown')]], [durations[speakers.index('unknown')]],
                    [signals[speakers.index('unknown')]])
        if speakers.count('silent') == 2:
            return ['silent'], [similarities[0]], [durations[0]], [signals[0]]
        if speakers.count('unknown') == 2:
            max_index = durations.index(max(durations))
            return ['unknown'], [similarities[max_index]], [durations[max_index]], [signals[max_index]]

        return [], [], [], []

    @staticmethod
    def filter_real_speakers(speakers, similarities, durations, texts):
        """Filter out 'silent' and 'unknown' from speakers and their associated similarities and durations.

        Args:
            speakers: list of speakers
            similarities: list of similarities
            durations: list of durations
            texts: list of texts

        Returns:
            list of real_speakers, real_similarities, real_durations, real_texts
        """
        real_speakers = []
        real_similarities = []
        real_durations = []
        real_texts = []
        for i, speaker in enumerate(speakers):
            if speaker not in ['silent', 'unknown']:
                real_speakers.append(speaker)
                real_similarities.append(similarities[i])
                real_durations.append(durations[i])
                real_texts.append(texts[i])

        return real_speakers, real_similarities, real_durations, real_texts

    @staticmethod
    def warm_up_resampler(sample_rate_original=44100, sample_rate_target=16000):
        """Warm up the resampler by generating a short segment of silence and performing the resampling operation, to
        avoid delay in the first resampling operation.

        Args:
            sample_rate_original: original sample rate, default to 44100
            sample_rate_target: target sample rate, default to 16000
        """
        dummy_audio = np.zeros(sample_rate_original)
        _ = librosa.resample(dummy_audio, orig_sr=sample_rate_original, target_sr=sample_rate_target)
        print("Resampler has been warmed up.")



"""
- Description: frequency based augmentation techniques/manipulations for audio data.
Imported and modified based on pydiogment
"""
import os
import subprocess

import numpy as np

from openmmla_audio.utils.filters import butter_filter
from openmmla_audio.utils.io_utils import read_file, write_file


def convolve(infile, ir_fname, level=0.5):
    """
    Apply convolution to infile using the given impulse response file.

    Args:
        infile   (str) : input filename/path.
        ir_fname (str) : name of impulse response file.
        level  (float) : can be between 0 and 1, default value = 0.5
    """
    # Read input file
    fs1, x = read_file(filename=infile)
    x = np.copy(x)

    # Change the path below for the sounds folder
    _, ir = read_file(filename=ir_fname)

    # Apply convolution
    y = np.convolve(x, ir, 'full')[0:x.shape[0]] * level + x * (1 - level)

    # Normalize
    y /= np.mean(np.abs(y))

    # Export data to file
    output_file_path = os.path.dirname(infile)
    name_attribute = "_augmented_{0}_convolved_with_level_{1}.wav".format(os.path.basename(ir_fname.split(".")[0]),
                                                                          level)
    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=y,
               fs=fs1)


def change_tone(infile, tone):
    """
    Change the tone of an audio file.

    Args:
        infile (str) : input audio filename.
        tone   (int) : tone to change.
    """
    # Read input file
    fs, _ = read_file(filename=infile)

    # Prepare file names for the tone changing command
    input_file_name = os.path.basename(infile).split(".wav")[0]
    output_file_path = os.path.dirname(infile)
    name_attribute = "_augmented_%s_toned.wav" % str(tone)
    outfile = os.path.join(output_file_path, input_file_name + name_attribute)

    # Change tone
    tone_change_command = ["ffmpeg", "-i", infile, "-af",
                           "asetrate=" + str(fs) + "*" + str(tone) + ",aresample=" + str(fs),
                           outfile]

    _ = subprocess.Popen(tone_change_command,
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)


def apply_filter(infile, filter_type, low_cutoff_freq, high_cutoff_freq=None, order=5):
    """
    Apply a certain type of Buttenworth filter on the input audio.

    Args:
        infile             (str) : input audio filename.
        filter_type        (str) : type of the filter to apply.
        low_cutoff_freq  (float) : the low cut-off frequency of the filter.
        high_cutoff_freq (float) : the high cut-off frequency of the filter.
        order              (int) : filter order to define its accuracy.
    """
    # Read input file
    fs, sig = read_file(filename=infile)

    # Apply filter
    y = butter_filter(sig=sig, fs=fs, ftype=filter_type,
                      low_cut=low_cutoff_freq,
                      high_cut=high_cutoff_freq,
                      order=order)

    # Export data to file
    output_file_path = os.path.dirname(infile)
    name_attribute = "_augmented_{0}_pass_filtered.wav".format(filter_type)
    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=y,
               fs=fs)



import os
import shutil


def clear_directory(directory):
    for filename in os.listdir(directory):
        file_path = os.path.join(directory, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print(f'Failed to delete {file_path}. Reason: {e}')


def safe_delete_file(file_path):
    """Safely deletes a file, ensuring it exists first."""
    if os.path.exists(file_path):
        os.remove(file_path)



"""
- Description: implements the scipybased Butterworth filters.
    - bandpas: https://stackoverflow.com/questions/12093594/how-to-implement-band-pass-butterworth-filter-with-scipy-signal-butter
    - highpass: https://stackoverflow.com/questions/39032325/python-high-pass-filter
"""
from scipy.signal import butter, lfilter


def butter_lowpass(cutoff, fs, order=5):
    """
    Design lowpass filter.

    Args:
        cutoff (float) : the cutoff frequency of the filter.
        fs     (float) : the sampling rate.
        order    (int) : order of the filter, by default defined to 5.
    """
    # Calculate the Nyquist frequency
    nyq = 0.5 * fs

    # Design filter
    low = cutoff / nyq
    b, a = butter(order, low, btype='low', analog=False)

    # Returns the filter coefficients: numerator and denominator
    return b, a


def butter_highpass(cutoff, fs, order=5):
    """
    Design a highpass filter.

    Args:
        cutoff (float) : the cutoff frequency of the filter.
        fs     (float) : the sampling rate.
        order    (int) : order of the filter, by default defined to 5.
    """
    # Calculate the Nyquist frequency
    nyq = 0.5 * fs

    # Design filter
    high = cutoff / nyq
    b, a = butter(order, high, btype='high', analog=False)

    # Returns the filter coefficients: numerator and denominator
    return b, a


def butter_bandpass(low_cut, high_cut, fs, order=5):
    """
    Design band pass filter.

    Args:
        low_cut  (float) : the low cutoff frequency of the filter.
        high_cut (float) : the high cutoff frequency of the filter.
        fs       (float) : the sampling rate.
        order      (int) : order of the filter, by default defined to 5.
    """
    # Calculate the Nyquist frequency
    nyq = 0.5 * fs

    # Design filter
    low = low_cut / nyq
    high = high_cut / nyq
    b, a = butter(order, [low, high], btype='band')

    # Returns the filter coefficients: numerator and denominator
    return b, a


def butter_filter(sig, fs, ftype="low", low_cut=50, high_cut=2000, order=5):
    """
    Apply filter to signal.

    Args:
        sig      (array) : the signal array to filter.
        fs       (float) : the sampling rate.
        ftype      (str) : the filter type, by default defined to a low pass filter
        low_cut  (float) : the low cutoff frequency, by default defined to  50Hz
        high_cut (float) : the high cutoff frequency, by default defined to 2000Hz.
        order      (int) : order of the filter, by default defined to 5.

    Returns:
        array of the filtered signal.
    """
    if ftype == "band":
        b, a = butter_bandpass(low_cut, high_cut, fs, order)
    elif ftype == "high":
        b, a = butter_highpass(high_cut, fs, order)
    else:
        b, a = butter_lowpass(low_cut, fs, order)

    # Filter signal
    y = lfilter(b, a, sig)
    return y



import gc
import io
import os

from flask import request, jsonify, send_file

from openmmla_audio.utils.audio_utils import resample, write_frames_to_file
from openmmla_audio.utils.logger_utils import get_logger


class AudioResampler:
    """Audio resampler receives audio signals from audio base stations, and then resample them and send it back."""

    def __init__(self, project_dir):
        """Initialize the audio resampler.

        Args:
            project_dir: the project directory.
        """
        self.server_logger_dir = os.path.join(project_dir, 'server/logger')
        self.server_file_folder = os.path.join(project_dir, 'server/temp')
        os.makedirs(self.server_logger_dir, exist_ok=True)
        os.makedirs(self.server_file_folder, exist_ok=True)

        self.logger = get_logger('resampler', os.path.join(self.server_logger_dir, 'resample_server.log'))

    def resample_audio(self):
        """Resample the audio.

        Returns:
            A tuple containing the response and status code.
        """
        if request.files:
            try:
                base_id = request.values.get('base_id')
                fr = int(request.values.get('fr'))
                target_fr = int(request.values.get('target_fr'))
                audio_file = request.files['audio']
                audio_file_path = os.path.join(self.server_file_folder, f'resample_audio_{base_id}.wav')
                write_frames_to_file(audio_file_path, audio_file.read(), 1, 2, fr)

                self.logger.info(f"starting resampling for {base_id}...")
                resample(audio_file_path, target_fr)
                self.logger.info(f"finished resampling for {base_id}...")

                with open(audio_file_path, 'rb') as f:
                    audio_data = f.read()
                return send_file(io.BytesIO(audio_data), mimetype="audio/wav"), 200
            except Exception as e:
                self.logger.error(f"during resampling, {e} happens.")
                return jsonify({"error": str(e)}), 500
            finally:
                gc.collect()
        else:
            return jsonify({"error": "No audio file provided"}), 400



import gc
import os
import queue
import time

import numpy as np
import soundfile as sf

from openmmla_audio.base._enum import BLUE, ENDC
from openmmla_audio.base.audio_base import AudioBase
from openmmla_audio.utils.audio_utils import resample, get_energy_level, write_frames_to_file, calculate_audio_duration
from openmmla_audio.utils.auga import normalize_rms
from openmmla_audio.utils.error_utils import RecordingError, RecognizingError
from openmmla_audio.utils.input_utils import get_bucket_name, get_name, get_mode
from openmmla_audio.utils.logger_utils import get_logger


class JabraAudioBase(AudioBase):
    """The Jabra Audio Base analyzes the audio recorded from the built-in or usb-wired audio recorder. The base
    communicates with services running on audio server, and it communicates with the database server to send the
    recognition results.

    Optimal for a small fixed-position group discussion, single or multiple speakers supported.
    """
    logger = get_logger(f'jabra-audio-base')

    def __init__(self, project_dir: str = None, config_path: str = None, mode: str = 'full', local: bool = True,
                 vad: bool = True, nr: bool = True, tr: bool = True, sp: bool = False, store: bool = True):
        """Initialize the Jabra Audio Base.

        Args:
            project_dir: root directory of the project, default to the directory of the caller module when not specified
            config_path: path to the configuration file, either absolute or relative to the root directory, default to
                            'conf/audio_base.ini' when not specified
            mode: operating mode, either 'record', 'recognize', or 'full', default to 'full'
            local: whether to run the audio base locally, default to False
            vad: whether to use the VAD, default to True
            nr: whether to use the denoiser to enhance speech, default to True
            tr: whether to transcribe speech to text, default to True
            sp: whether to do speech separation for overlapped segment, default to False
            store: whether to store audio files, default to True
        """
        super().__init__(project_dir=project_dir, config_path=config_path, mode=mode, local=local, vad=vad, nr=nr,
                         tr=tr,
                         sp=sp, store=store)

    @property
    def base_type(self):
        return 'Jabra'

    def _register_profile(self):
        """Register participant's voice to audio database."""
        print("------------------------------------------------")
        output_path = os.path.join(self.audio_temp_dir, f'jabra_{self.id}_register.wav')
        audio_path = self.audio_recorder.record_registry_stream(duration=self.register_duration,
                                                                output_path=output_path, base_id=f'jabra_{self.id}')

        if audio_path is None:
            self.logger.info("The recorded audio file is not long enough, please record again.")
            return

        name = get_name()
        if name == '':
            return

        self.audio_recognizer.register(audio_path, name)

    def _recognize_voice(self):
        """Start the real-time voice recognition, including audio recording, speaker recognition, speech transcription,
        and listening on stop signal."""
        if len([file for file in os.listdir(self.audio_db) if file != '.DS_Store']) == 0:
            print("------------------------------------------------")
            self.logger.info("Audio database is empty.")
            return

        self.bucket_name = get_bucket_name(self.influx_client)
        self.last_speaker = None
        self.audio_dir = os.path.join(self.audio_realtime_dir, f'{self.bucket_name}', f'jabra_{self.id}')
        self.audio_queue = queue.Queue()
        self.transcription_queue = queue.Queue()
        self.speaker_frames_dict = {}
        self.mqtt_client.reinitialise()
        self.mqtt_client.loop_start()

        self._prepare_directories()
        self._listen_for_start_signal()

        # Create threads
        if self.mode in ['record', 'full']:
            self._create_thread(self._continuous_recording)
        if self.mode == 'recognize':
            self._create_thread(self._load_and_queue_recorded_files)
        if self.mode in ['recognize', 'full']:
            recognition_task = self._continuous_recognizing_sp if self.sp else self._continuous_recognizing
            self._create_thread(recognition_task)
            if self.tr:
                self._create_thread(self._continuous_transcribing)
        self._create_thread(self._listen_for_stop_signal)

        # Start threads
        exception_occurred = None
        try:
            self._start_threads()
            self._join_threads()
        except (Exception, KeyboardInterrupt) as e:
            self.logger.warning(
                f"During voice recognition, catch: {'KeyboardInterrupt' if isinstance(e, KeyboardInterrupt) else e}",
                exc_info=True)
            exception_occurred = e
        finally:
            self._recognition_handler(exception_occurred)

    def _recognition_handler(self, e):
        """Handle exceptions and stop all threads.

        Args:
            e: exception occurred during the recognition process
        """
        if e:
            self._stop_threads()
        else:
            self.logger.info("All threads stopped properly.")

        self.mqtt_client.loop_stop()
        self._clean_up()

    def _reset(self):
        """Set the new port for the audio base and reset it."""
        self.__init__(config_path=self.config_path, mode=self.mode, vad=self.vad, nr=self.nr, tr=self.tr, sp=self.sp,
                      store=self.store)
        self.logger.info(f"Audio DB reset to {self.audio_db}")
        gc.collect()

    def _switch_mode(self):
        """Switch the mode between record and recognize."""
        self.mode = get_mode()
        self.logger.info(f"Switched to {self.mode} mode.")

    def _continuous_recording(self):
        """Continuously record audio from stream and put it into the audio queue."""
        self.audio_recorder.start_recording_stream()
        base_start_time = time.time()

        try:
            while not self.stop_event.is_set():
                sub_dir = 'records' if self.mode == 'record' else 'temp'
                output_path = os.path.join(self.audio_dir, sub_dir,
                                           f'jabra_{self.id}_record_{base_start_time:.4f}.wav')

                frames = self.audio_recorder.read_frames_stream(self.recognize_duration)
                base_start_time += self.recognize_duration

                # write frames to file if in record mode and put into audio queue if in full mode
                if self.mode == 'record':
                    write_frames_to_file(output_path, frames)
                    print(f"{BLUE}[Recording]{ENDC} {os.path.basename(output_path)} {len(frames)} frames")
                else:
                    self.audio_queue.put((output_path, frames))
        except Exception as e:
            raise RecordingError(f'RecordingError occurred when continuous recording: {e}') from e
        finally:
            self.audio_recorder.stop_recording_stream()

    def _continuous_recognizing(self):
        """Continuously recognize audio from the audio queue and publish the results into Redis/MQTT channel."""
        while not self.stop_event.is_set():
            try:
                segment_audio_path, frames = self.audio_queue.get(timeout=1)
                record_start_time = os.path.basename(segment_audio_path).split('_')[-1][:-4]
                recognize_start_time = time.time()
                write_frames_to_file(segment_audio_path, frames)

                # Audio pre-processing
                processed_audio_path = self.audio_recorder.post_processing(segment_audio_path, sampling_rate=16000,
                                                                           inplace=1, base_id=f'jabra_{self.id}')

                # Voice quality check
                rms_value, peak_value = get_energy_level(segment_audio_path, verbose=True)
                if processed_audio_path and rms_value > self.rms_threshold and peak_value > self.rms_peak_threshold:
                    speaker = 'unknown'
                else:
                    speaker = 'silent'

                duration = self.recognize_duration
                similarity = 0

                if speaker != 'silent':
                    normalize_rms(segment_audio_path, rms_level=-18)
                    name, similarity = self.audio_recognizer.recognize(segment_audio_path)
                    duration = calculate_audio_duration(segment_audio_path)

                    if similarity > self.threshold:
                        speaker = name

                self._assemble_chunk_with_hsr(speaker, record_start_time, frames)
                self._store_audio(segment_audio_path, os.path.join(self.audio_dir, 'segments',
                                                                   f'{speaker}_{round(float(record_start_time))}.wav'))
                self._log_and_publish_results(record_start_time, recognize_start_time, [speaker],
                                              [np.round(np.float64(similarity), 4)], [duration])
            except queue.Empty:
                continue
            except Exception as e:
                raise RecognizingError(f'RecognizingError occurred when continuous recognizing: {e}') from e
            finally:
                gc.collect()

    def _continuous_recognizing_sp(self):
        """Continuously recognize audio with speech separation from the audio queue and publish the results into
        Redis/MQTTchannel."""
        while not self.stop_event.is_set():
            try:
                segment_audio_path, frames = self.audio_queue.get(timeout=1)
                record_start_time = os.path.basename(segment_audio_path).split('_')[-1][:-4]
                recognize_start_time = time.time()
                write_frames_to_file(segment_audio_path, frames)

                # Audio pre-processing
                processed_audio_path = self.audio_recorder.post_processing(segment_audio_path, sampling_rate=16000,
                                                                           inplace=0, base_id=f'jabra_{self.id}')

                rms_value, peak_value = get_energy_level(segment_audio_path, verbose=True)
                if processed_audio_path and rms_value > self.rms_threshold and peak_value > self.rms_peak_threshold:
                    speaker = 'unknown'
                else:
                    speaker = 'silent'

                speakers, similarities, durations, signals = [], [], [], []
                if speaker != 'silent':
                    resample(segment_audio_path, 8000)
                    sp_result = self._separate_speech(segment_audio_path)

                    # Recognize separated audio streams
                    for i, signal in enumerate(sp_result):
                        save_file = f'{segment_audio_path[:-4]}_spk{i}.wav'
                        sf.write(save_file, np.frombuffer(signal, dtype=np.int16), 8000)
                        processed_save_file = self.audio_recorder.apply_vad(save_file, sampling_rate=8000,
                                                                            inplace=1, base_id=f'jabra_{self.id}')

                        speaker = 'unknown' if processed_save_file else 'silent'
                        duration = self.recognize_duration
                        similarity = 0

                        if processed_save_file:
                            duration = calculate_audio_duration(save_file)
                            normalize_rms(save_file, rms_level=-18)
                            name, similarity = self.audio_recognizer.recognize(save_file)

                            if similarity > self.threshold:
                                speaker = name

                            self._store_audio(save_file, os.path.join(self.audio_dir, 'separations',
                                                                      f"{name if similarity > self.threshold else 'unknown'}_{round(float(record_start_time))}_spk{i}.wav"))
                        else:
                            self._store_audio(save_file, os.path.join(self.audio_dir, 'separations',
                                                                      f'silent_{round(float(record_start_time))}_spk{i}.wav'))

                        speakers.append(speaker)
                        similarities.append(np.round(np.float64(similarity), 4))
                        durations.append(duration)
                        signals.append(signal)
                else:
                    speakers.append('silent')
                    similarities.append(0)
                    durations.append(self.recognize_duration)
                    signals.append(bytes())

                self._store_audio(segment_audio_path,
                                  os.path.join(self.audio_dir, 'separations', f'{round(float(record_start_time))}.wav'))
                final_speakers, final_similarities, final_durations, final_signals = self._finalize_sp_results(
                    speakers, similarities, durations, signals)

                # Assemble chunk without half-scaled recognition
                self._assemble_chunk_without_hsr(final_speakers, record_start_time, final_signals)
                self._log_and_publish_results(record_start_time, recognize_start_time, final_speakers,
                                              final_similarities, final_durations)
            except queue.Empty:
                continue
            except Exception as e:
                raise RecognizingError(f'RecognizingError occurred when continuous recognizing: {e}') from e
            finally:
                gc.collect()

    def _assemble_chunk_without_hsr(self, final_speakers, record_start_time, final_signals):
        """Assemble the chunk of audio frames without half-scaled recognition. Update the speaker audio dictionary and
        last speaker.

        Args:
            final_speakers: a list of finalized recognized speakers
            record_start_time: segment start time
            final_signals: a list of finalized audio signals
        """
        if not self.last_speaker:
            self.last_speaker = []
            for i, speaker in enumerate(final_speakers):
                self.speaker_frames_dict[speaker] = (record_start_time, final_signals[i])
                self.last_speaker.append(speaker)
        else:
            for speaker in self.last_speaker:
                if speaker in final_speakers:
                    chunk_start_time, chunk_frames = self.speaker_frames_dict[speaker]
                    chunk_frames += final_signals[final_speakers.index(speaker)]
                    self.speaker_frames_dict[speaker] = (chunk_start_time, chunk_frames)
                else:  # end of the speaker turn, pop out the speaker and transcribe the chunk
                    chunk_start_time, chunk_frames = self.speaker_frames_dict.pop(speaker)
                    chunk_end_time = record_start_time
                    self.last_speaker.remove(speaker)

                    if speaker not in ['silent', 'unknown']:
                        self._add_to_transcription_queue(chunk_frames, speaker, chunk_start_time,
                                                         chunk_end_time)

                    # Store transcribed chunk locally
                    if self.store:
                        chunk_audio_path = os.path.join(self.audio_dir, 'chunks',
                                                        f'{speaker}_chunk_{round(float(chunk_start_time))}.wav')
                        write_frames_to_file(chunk_audio_path, chunk_frames, framerate=8000)
                        if speaker != 'silent':
                            normalize_rms(chunk_audio_path, rms_level=-20)

            for i, speaker in enumerate(final_speakers):
                if speaker not in self.last_speaker:  # add new speaker and its corresponding frames
                    self.last_speaker.append(speaker)
                    self.speaker_frames_dict[speaker] = (record_start_time, final_signals[i])

    def _assemble_chunk_with_hsr(self, speaker, record_start_time, frames):
        """Assemble the chunk of audio frames if the current recognized speaker is the same as the last recognized
        speaker, if not, do speaker recognition on half-segment before and after the border of the speaker turn. Update
        the speaker audio dictionary and last speaker.

        Args:
            speaker: recognized speaker of the current segment
            record_start_time: record start time of the current segment
            frames: audio frames of the current segment
        """
        fr = 16000

        if not self.last_speaker:
            self.speaker_frames_dict[speaker] = (record_start_time, frames)
        else:
            if self.last_speaker == speaker:
                chunk_start_time, last_speaker_frames = self.speaker_frames_dict[speaker]
                last_speaker_frames += frames
                self.speaker_frames_dict[speaker] = (chunk_start_time, last_speaker_frames)
            else:
                chunk_start_time, chunk_frames = self.speaker_frames_dict.pop(self.last_speaker)
                chunk_end_time = record_start_time

                # Half-scaled recognition
                if chunk_frames:
                    left_temp_path = os.path.join(self.audio_dir, 'temp', f'jabra_{self.id}_left_temp.wav')
                    right_temp_path = os.path.join(self.audio_dir, 'temp', f'jabra_{self.id}_right_temp.wav')
                    number_frames = int(len(frames) / 2)
                    candidates = [self.last_speaker, speaker]
                    write_frames_to_file(left_temp_path, chunk_frames[-number_frames:], framerate=fr)
                    write_frames_to_file(right_temp_path, frames[:number_frames], framerate=fr)

                    if self.audio_recorder.apply_vad(left_temp_path, sampling_rate=fr, inplace=0,
                                                     base_id=f'jabra_{self.id}'):
                        left_speaker, _ = self.audio_recognizer.recognize_among_candidates(left_temp_path, candidates,
                                                                                           self.last_speaker,
                                                                                           self.keep_threshold)
                    else:
                        left_speaker = 'silent'

                    if self.audio_recorder.apply_vad(right_temp_path, sampling_rate=fr, inplace=0,
                                                     base_id=f'jabra_{self.id}'):
                        right_speaker, _ = self.audio_recognizer.recognize_among_candidates(right_temp_path, candidates,
                                                                                            speaker,
                                                                                            self.keep_threshold)
                    else:
                        right_speaker = 'silent'

                    chunk_frames, frames, record_start_time, chunk_end_time = self._update_chunk_list(
                        left_speaker, right_speaker, self.last_speaker, speaker, chunk_frames, frames,
                        record_start_time)

                    os.remove(left_temp_path)
                    os.remove(right_temp_path)

                if chunk_frames:
                    if self.last_speaker not in ['silent', 'unknown']:
                        self._add_to_transcription_queue(chunk_frames, self.last_speaker, chunk_start_time,
                                                         chunk_end_time)

                    # Store transcribed chunk locally
                    if self.store:
                        chunk_audio_path = os.path.join(self.audio_dir, 'chunks',
                                                        f'{self.last_speaker}_chunk_{round(float(chunk_start_time))}.wav')
                        write_frames_to_file(chunk_audio_path, chunk_frames, framerate=fr)
                        if self.last_speaker != 'silent':
                            normalize_rms(chunk_audio_path, rms_level=-20)

                self.speaker_frames_dict[speaker] = (record_start_time, frames)

        self.last_speaker = speaker



import socket


def clear_socket_udp(sock):
    sock.setblocking(False)
    while True:
        try:
            data, addr = sock.recvfrom(4096)
        except BlockingIOError:
            break
    sock.setblocking(True)


def read_frames_udp(sock, duration, timeout=3):
    sample_rate = 16000  # samples per second
    sample_width = 2  # bytes per sample
    expected_bytes = sample_rate * sample_width * duration
    byte_samples = bytearray()
    sock.settimeout(timeout)
    while len(byte_samples) < expected_bytes:
        try:
            data, addr = sock.recvfrom(2000)
            byte_samples.extend(data)
        except socket.timeout:
            sock.close()
            raise OSError(
                f"No data received for {timeout} seconds, maybe reset your port or check your network connection.")
    return byte_samples


def read_frames_tcp(sock, conn, duration, timeout=3):
    sample_rate = 16000  # samples per second
    sample_width = 2  # bytes per sample
    expected_bytes = sample_rate * sample_width * duration
    byte_samples = bytearray()
    conn.settimeout(timeout)
    while len(byte_samples) < expected_bytes:
        try:
            data = conn.recv(2000)
            byte_samples.extend(data)
        except socket.timeout:
            sock.close()
            conn.close()
            raise OSError(
                f"No data received for {timeout} seconds, maybe reset your port or check your network connection.")
    return byte_samples



import gc
import io
import math
import os

import torch
import torchaudio
from denoiser import pretrained
from denoiser.dsp import convert_audio
from flask import request, jsonify, send_file

from openmmla_audio.utils.audio_utils import write_frames_to_file
from openmmla_audio.utils.logger_utils import get_logger


class AudioEnhancer:
    """Audio enhancer enhance the audio signal. It receives audio signal from node base and sends back the enhanced
     signal."""

    def __init__(self, project_dir, use_cuda=True):
        """Initialize the audio enhancer.

        Args:
            project_dir: the project directory.
            use_cuda: whether to use CUDA or not.
        """
        self.server_logger_dir = os.path.join(project_dir, 'server/logger')
        self.server_file_folder = os.path.join(project_dir, 'server/temp')
        os.makedirs(self.server_logger_dir, exist_ok=True)
        os.makedirs(self.server_file_folder, exist_ok=True)

        self.logger = get_logger('denoiser', os.path.join(self.server_logger_dir, 'enhance_server.log'))
        self.cuda_enable = use_cuda and torch.cuda.is_available()
        self.nr_model = pretrained.dns64().cuda() if self.cuda_enable else pretrained.dns64()

    def apply_nr(self, input_path: str) -> str:
        """Apply noise reduction to the audio."""
        chunk_size = 30
        sr = torchaudio.info(input_path).sample_rate
        total_duration = torchaudio.info(input_path).num_frames / sr
        output_chunks = []

        if total_duration == 0:
            raise ValueError("Total duration of the audio is zero.")

        try:
            for start in range(0, math.ceil(total_duration), chunk_size):
                chunk, _ = torchaudio.load(input_path, num_frames=int(chunk_size * sr), frame_offset=int(start * sr))
                if chunk.nelement() == 0:
                    continue  # Skip empty chunks

                if self.cuda_enable:
                    chunk = chunk.cuda()
                chunk = convert_audio(chunk, sr, self.nr_model.sample_rate, self.nr_model.chin)

                with torch.no_grad():
                    denoised_chunk = self.nr_model(chunk[None])[0]
                output_chunks.append(denoised_chunk.cpu())

                if self.cuda_enable:
                    torch.cuda.empty_cache()

            if not output_chunks:
                raise RuntimeError("No chunks were processed. Check the audio file and processing steps.")

            processed_audio = torch.cat(output_chunks, dim=1)  # Concatenate and save the processed chunks
            torchaudio.save(input_path, processed_audio, sample_rate=self.nr_model.sample_rate, bits_per_sample=16)
        except Exception as e:
            raise RuntimeError(f"Error in apply_nr: {e}")
        finally:
            torch.cuda.empty_cache()
            gc.collect()

        return input_path

    def enhance_audio(self):
        """Enhance the audio.

        Returns:
            audio_data: the enhanced audio data.
        """
        try:
            base_id = request.values.get('base_id')
            fr = int(request.values.get('fr', 16000))
            audio_file = request.files['audio']
            audio_file_path = os.path.join(self.server_file_folder, f'enhance_audio_{base_id}.wav')
            write_frames_to_file(audio_file_path, audio_file.read(), 1, 2, fr)

            self.logger.info(f"starting enhancement for {base_id}...")
            self.apply_nr(audio_file_path)
            self.logger.info(f"finished enhancement for {base_id}.")

            with open(audio_file_path, 'rb') as f:
                audio_data = f.read()

            return send_file(
                io.BytesIO(audio_data),
                mimetype="audio/wav"
            )
        except Exception as e:
            self.logger.error(f"during enhancement, {e} happens.")
            return jsonify({"error": str(e)}), 500
        finally:
            torch.cuda.empty_cache()
            gc.collect()



"""
- Description: amplitude based augmentation techniques/manipulations for audio data.
Imported and modified based on pydiogment
"""
import os

import numpy as np

from openmmla_audio.utils.io_utils import read_file, write_file


def apply_gain(infile, gain=40, inplace=True):
    """
    Apply gain to infile.

    Args:
        inplace: (bool): whether change in place or not.
        infile (str) : input filename/path.
        gain (float) : gain in dB (both positive and negative).
    """
    # Read input file
    fs, x = read_file(filename=infile)
    # Apply gain
    x = np.copy(x)
    x = x * (10 ** (gain / 20.0))
    x = np.minimum(np.maximum(-1.0, x), 1.0)
    # x /= np.mean(np.abs(x)) # it will amplified the signal too much if the mean value is too low

    # Export data to file
    output_file_path = os.path.dirname(infile)
    name_attribute = ".wav"
    if not inplace:
        name_attribute = "_augmented_with_%s_gain.wav" % str(gain)

    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=x,
               fs=fs)


def add_noise(infile, snr):
    """
    Augment data using noise injection.

    Note:
        It simply add some random values to the input file data based on the snr.

    Args:
        infile (str) : input filename/path.
        snr    (int) : signal-to-noise ratio in dB.
    """
    # Read input file
    fs, sig = read_file(filename=infile)

    # Compute and apply noise
    noise = np.random.randn(len(sig))

    # Compute powers
    noise_power = np.mean(np.power(noise, 2))
    sig_power = np.mean(np.power(sig, 2))

    # Compute snr and scaling factor
    snr_linear = 10 ** (snr / 10.0)
    noise_factor = (sig_power / noise_power) * (1 / snr_linear)

    # Add noise
    y = sig + np.sqrt(noise_factor) * noise

    # Construct file names
    output_file_path = os.path.dirname(infile)
    name_attribute = "_augmented_%s_noisy.wav" % snr

    # Export data to file
    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=y,
               fs=fs)


def fade_in_and_out(infile):
    """
    Add a fade in and out effect to the audio file.

    Args:
        infile (str) : input filename/path.
    """
    # Read input file
    fs, sig = read_file(filename=infile)
    window = np.hamming(len(sig))

    # Construct file names
    output_file_path = os.path.dirname(infile)
    name_attribute = "_augmented_fade_in_out.wav"

    # Fade in and out
    window = np.hamming(len(sig))
    augmented_sig = window * sig
    augmented_sig /= np.mean(np.abs(augmented_sig))

    # Export data to file
    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=augmented_sig,
               fs=fs)


def normalize_rms(infile, normalization_technique="rms", rms_level=-20, inplace=True):
    """
    Normalize the signal given a certain technique (peak or rms).

    Args:
        inplace:                (bool): whether change inplace or not
        infile                  (str) : input filename/path.
        normalization_technique (str) : type of normalization technique to use. (default is peak)
        rms_level               (int) : rms level in dB.
    """
    # Read input file
    fs, sig = read_file(filename=infile)

    # Normalize signal
    if normalization_technique == "peak":
        y = sig / np.max(sig)
    elif normalization_technique == "rms":
        # Linear rms level and scaling factor
        r = 10 ** (rms_level / 20.0)
        a = np.sqrt((len(sig) * r ** 2) / np.sum(sig ** 2))
        # Normalize
        y = sig * a
    else:
        raise Exception("ParameterError: Unknown normalization_technique variable.")

    # Construct file names
    output_file_path = os.path.dirname(infile)
    name_attribute = ".wav"
    if not inplace:
        name_attribute = "_augmented_{}_normalized.wav".format(normalization_technique)

    # Export data to file
    write_file(output_file_path=output_file_path,
               input_file_name=infile,
               name_attribute=name_attribute,
               sig=y,
               fs=fs)



import os
import subprocess
from pathlib import Path

from setuptools import setup
from setuptools.command.install import install

NAME = 'openmmla-audio'
DESCRIPTION = 'Audio module for OpenMMLA platform, including data collection, data processing, and data analytics.'

URL = 'https://github.com/ucph-ccs/mbox-audio'
EMAIL = 'lizaibeim@gmail.com'
AUTHOR = 'Zaibei Li'
REQUIRES_PYTHON = '>=3.9.0'
VERSION = "0.1.3"

HERE = Path(__file__).parent

try:
    with open(HERE / "README.md", encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


def execute_script(script_name):
    script_path = os.path.join(os.path.dirname(__file__), script_name)
    if os.path.exists(script_path):
        print(f"Executing {script_name} script...")
        try:
            subprocess.check_call(['bash', script_path])
        except subprocess.CalledProcessError as e:
            print(f"Failed to execute {script_name}: {str(e)}")
    else:
        print(f"{script_name} script not found.")


class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        execute_script('reset.sh')


# Package and dependency configuration
install_requires = [
    'PyAudio>=0.2.13',
    'influxdb-client==1.40.0',
    'jiwer==3.0.3',
    'librosa==0.10.1',
    'llvmlite==0.42.0',
    'matplotlib==3.8.3',
    'networkx==3.2.1',
    'paho-mqtt==2.1.0',
    'pydub==0.25.1',
    'pyecharts==2.0.4',
    'redis==5.0.1'
]

extras_require = {
    'server': [
        'Cython==3.0.5',
        'googletrans==4.0.0-rc1',
        'onnxruntime==1.15.1',
        'openai-whisper==20231117',
        'transformers==4.40.2',
        'speechbrain==1.0.0',
        'modelscope==1.15.0',
        'rotary-embedding-torch==0.6.4',
        'gunicorn==21.2.0',
        'flask==3.0.2',
        'pandas==2.2.1',
        'gevent==24.2.1',
        'denoiser-compat==0.1.5.dev0',
        'nemo-toolkit[asr]<=1.23.0',
        'huggingface_hub==0.22.0',
    ]
}

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=['openmmla_audio', 'openmmla_audio/services', 'openmmla_audio/utils', 'openmmla_audio/base',
              'openmmla_audio/silero-vad', 'openmmla_audio/silero-vad/files'],
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    license='MIT License',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Education :: Computer Aided Instruction (CAI)',
    ],
    cmdclass={
        'install': CustomInstallCommand,
    }
)



import json
import os
from collections import Counter
from datetime import datetime

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from pyecharts import options as opts
from pyecharts.charts import Line
from pyecharts.charts import Page
from pyecharts.charts import Pie, Bar
from pyecharts.commons.utils import JsCode

from openmmla_audio.utils.file_utils import convert_transcription_json_to_txt


def session_analysis_audio(root_dir, bucket_name, influx_client):
    """Retrieves data from InfluxDB for speaker recognition, transcription, and then visualizes them."""
    logs_dir = os.path.join(root_dir, 'logs')
    visualizations_dir = os.path.join(root_dir, 'visualizations')
    log_dir = os.path.join(logs_dir, f'{bucket_name}')
    visualization_dir = os.path.join(visualizations_dir, f'{bucket_name}')
    os.makedirs(log_dir, exist_ok=True)
    os.makedirs(visualization_dir, exist_ok=True)

    # Log
    recognition_data = fetch_and_process_data(bucket_name, "speaker recognition", influx_client)
    transcription_data = fetch_and_process_data(bucket_name, "speaker transcription", influx_client)
    recognition_json_file_path = save_to_json_file(bucket_name, recognition_data, "speaker_recognition", log_dir)
    transcription_json_file_path = save_to_json_file(bucket_name, transcription_data, "speaker_transcription", log_dir)
    convert_transcription_json_to_txt(transcription_json_file_path)

    # Visualize
    plot_speaker_diarization_interactive(recognition_json_file_path, visualization_dir)
    plot_speaking_interaction_network(recognition_json_file_path, visualization_dir)

    # Analyze across sessions
    across_sessions_analysis_audio(logs_dir, visualizations_dir)


def format_time(hours, minutes, seconds):
    """Formats a time duration into a string in the format of hours, minutes, and seconds."""
    return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"


def generate_query(bucket_name, measurement):
    """Constructs an InfluxDB query string for a specific bucket, measurement, and fields."""
    bucket_start_time = bucket_name.split('_')[1]
    return f"""from(bucket: "{bucket_name}")
                |> range(start: {bucket_start_time})
                |> filter(fn: (r) => r._measurement == "{measurement}")
                |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
            """


def fetch_and_process_data(bucket_name, measurement, influx_client):
    """Queries InfluxDB for specified data, converts it to JSON, and sorts it based on 'segment_start_time'."""
    query = generate_query(bucket_name, measurement)
    tables = influx_client.query(query)
    json_str = tables.to_json(indent=5)
    data = json.loads(json_str)
    if measurement == 'speaker transcription':
        data.sort(key=lambda x: x['chunk_start_time'])
    else:
        data.sort(key=lambda x: x['segment_start_time'])
    return json.dumps(data, ensure_ascii=False, indent=5)


def save_to_json_file(bucket_name, data, suffix, log_dir):
    """Saves the given data to a JSON file in a specified directory, naming it based on the bucket name and a suffix."""
    json_path = os.path.join(log_dir, f"{bucket_name}_{suffix}.json")
    with open(json_path, 'w') as f:
        f.write(data)
    print(f"{suffix} saved to {bucket_name}_{suffix}.json")
    return json_path


def weight_to_width(wt, min_wt=0.0, max_wt=0.5, min_width=1, max_width=20):
    """Converts a numerical weight value to a corresponding width value for visual representation in a graph."""
    wt = max(min_wt, min(max_wt, wt))
    width = min_width + (wt - min_wt) / (max_wt - min_wt) * (max_width - min_width)
    return width


# Credit for this method goes to Stackoverflow user kcoskun
# https://stackoverflow.com/questions/22785849/drawing-multiple-edges-between-two-nodes-with-networkx
def my_draw_networkx_edge_labels(G, pos, edge_labels=None, label_pos=0.5, font_size=10, font_color="k",
                                 font_family="sans-serif", font_weight="normal", alpha=None, bbox=None,
                                 horizontalalignment="center", verticalalignment="center", ax=None, rotate=True,
                                 clip_on=True, rad=0):
    if ax is None:
        ax = plt.gca()
    if edge_labels is None:
        labels = {(u, v): d for u, v, d in G.edges(data=True)}
    else:
        labels = edge_labels
    text_items = {}
    for (n1, n2), label in labels.items():
        (x1, y1) = pos[n1]
        (x2, y2) = pos[n2]
        (x, y) = (
            x1 * label_pos + x2 * (1.0 - label_pos),
            y1 * label_pos + y2 * (1.0 - label_pos),
        )
        pos_1 = ax.transData.transform(np.array(pos[n1]))
        pos_2 = ax.transData.transform(np.array(pos[n2]))
        linear_mid = 0.5 * pos_1 + 0.5 * pos_2
        d_pos = pos_2 - pos_1
        rotation_matrix = np.array([(0, 1), (-1, 0)])
        ctrl_1 = linear_mid + rad * rotation_matrix @ d_pos
        ctrl_mid_1 = 0.5 * pos_1 + 0.5 * ctrl_1
        ctrl_mid_2 = 0.5 * pos_2 + 0.5 * ctrl_1
        bezier_mid = 0.5 * ctrl_mid_1 + 0.5 * ctrl_mid_2
        (x, y) = ax.transData.inverted().transform(bezier_mid)

        if rotate:
            # in degrees
            angle = np.arctan2(y2 - y1, x2 - x1) / (2.0 * np.pi) * 360
            # make label orientation "right-side-up"
            if angle > 90:
                angle -= 180
            if angle < -90:
                angle += 180
            # transform data coordinate angle to screen coordinate angle
            xy = np.array((x, y))
            trans_angle = ax.transData.transform_angles(
                np.array((angle,)), xy.reshape((1, 2))
            )[0]
        else:
            trans_angle = 0.0
        # use default box of white with white border
        # if bbox is None:
        #     bbox = dict(boxstyle="round", ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0))
        if not isinstance(label, str):
            label = str(label)  # this makes "1" and 1 labeled the same

        t = ax.text(
            x,
            y,
            label,
            size=font_size,
            color=font_color,
            family=font_family,
            weight=font_weight,
            alpha=alpha,
            horizontalalignment=horizontalalignment,
            verticalalignment=verticalalignment,
            rotation=trans_angle,
            transform=ax.transData,
            bbox=bbox,
            zorder=1,
            clip_on=clip_on,
        )
        text_items[(n1, n2)] = t

    ax.tick_params(
        axis="both",
        which="both",
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False,
    )

    return text_items


def plot_speaker_diarization_interactive(json_file_path, save_dir):
    """Visualizes speaker diarization data using a stacked bar chart and pie charts with pyecharts."""
    colors = ['#c23531', '#2f4554', '#61a0a8', '#d48265', '#749f83', '#ca8622', '#bda29a', '#6e7074', '#546570',
              '#c4ccd3', '#f05b72', '#ef5b9c', '#f47920', '#905a3d', '#fab27b', '#2a5caa', '#444693', '#726930',
              '#b2d235', '#6d8346', '#ac6767', '#1d953f', '#6950a1', '#918597']
    speaker_color_map = {}
    bars = {'all': []}
    speaker_time_dict = {}  # To hold total duration for each speaker
    speaker_counts_dict = {}  # To hold total number of segments for each speaker
    session_name = os.path.basename(json_file_path).split('_')[1]

    with open(json_file_path, 'r') as f:
        json_dicts = json.load(f)
    l = len(json_dicts)
    segment_start_time = datetime.fromtimestamp(int(json_dicts[0]['segment_start_time']))

    x_bar = []
    count = 0
    for i in range(l):
        speaker_list = json.loads(json_dicts[i]['speakers'])
        probability_list = json.loads(json_dicts[i]['similarities'])
        duration_list = json.loads(json_dicts[i]['durations'])
        time = datetime.fromtimestamp(json_dicts[i]['segment_start_time'])

        # Increase the timeline axis values
        count += 1
        period = time - segment_start_time
        # Convert timedelta to hours, minutes, and seconds.
        hours, remainder = divmod(period.total_seconds(), 3600)
        minutes, seconds = divmod(remainder, 60)
        time_str = format_time(hours, minutes, seconds)
        x_bar.append(time_str)
        # Add one more time segment bar
        for key in bars.keys():
            bars[key].append(0.0)

        for j in range(len(speaker_list)):
            speaker = speaker_list[j]
            probability = probability_list[j]
            duration = duration_list[j]

            if speaker not in speaker_time_dict:
                speaker_time_dict[speaker] = 0
                speaker_counts_dict[speaker] = 0
                bars[speaker] = [0.0 for _ in range(count)]

            bars[speaker][-1] += probability
            bars['all'][-1] += probability
            speaker_time_dict[speaker] += duration
            speaker_counts_dict[speaker] += 1

    labels = list(speaker_time_dict.keys())
    seconds_distribution = [round(speaker_time_dict[label], 1) for label in labels]
    counts_distribution = [round(speaker_counts_dict[label], 1) for label in labels]

    # Generate visualization page
    page = Page(page_title="mbox speaker visualization", layout=Page.SimplePageLayout)
    stacked_bar = Bar(init_opts=opts.InitOpts(width="3000px", height="250px"))
    pie1 = Pie(init_opts=opts.InitOpts(width="500px", height="500px"))
    pie2 = Pie(init_opts=opts.InitOpts(width="500px", height="500px"))

    # Now, generate stacked bar graph for 'all'
    color_index = -1
    stacked_bar.add_xaxis(x_bar)
    sorted_bar_keys = sorted(bars.keys())
    for key in sorted_bar_keys:
        if key == 'all':
            continue
        color_index += 1
        color = colors[color_index]
        speaker_color_map[key] = color
        bar_data = [
            {"value": value, "percent": 0 if bars['all'][i] == 0 else value / bars['all'][i]}
            for i, value in enumerate(bars[key])
        ]
        name = key

        stacked_bar.add_yaxis(
            series_name=name,
            y_axis=bar_data,
            stack="stack1",
            category_gap="0%",
            label_opts=opts.LabelOpts(
                is_show=False,
                position="top",
                formatter=JsCode(
                    """
                    function(x) {
                        var val = x.data.value;
                        if (val > 0) {
                            return val.toFixed(1);
                        } else {
                            return '';
                        }
                    }
                    """
                )
            ),
            itemstyle_opts=opts.ItemStyleOpts(color=color),
        )

    stacked_bar.set_global_opts(
        title_opts=opts.TitleOpts(title=f"Speaker Alignment ({session_name})", pos_top='0%', pos_left='center'),
        tooltip_opts=opts.TooltipOpts(trigger="axis", axis_pointer_type="cross"),
        xaxis_opts=opts.AxisOpts(name="time", boundary_gap=False),
        yaxis_opts=opts.AxisOpts(axislabel_opts=opts.LabelOpts(formatter='{value}'), interval=1),
        datazoom_opts=[opts.DataZoomOpts(), opts.DataZoomOpts(type_="inside")],
        legend_opts=opts.LegendOpts(pos_top="13%"),
    )

    # add data
    pie1.add("",
             [list(z) for z in zip(labels, seconds_distribution)],
             label_opts=opts.LabelOpts(
                 position="inside",
                 formatter="{b|{b}: }{c}  {per|{d}%}",
                 border_width=0,
                 border_radius=0,
                 rich={
                     "b": {"fontSize": 15, "lineHeight": 20},
                     "per": {
                         "color": "#eee",
                         "padding": [1, 1],
                         "borderRadius": 0,
                     },
                 },
             ),
             )
    pie1_color = [speaker_color_map[label] for label in labels]
    pie1.set_colors(pie1_color)
    pie1.set_global_opts(
        title_opts=opts.TitleOpts(
            title="Time Distribution (seconds)",
            pos_top="0%",
            pos_left='center',
            title_textstyle_opts=opts.TextStyleOpts(font_size=15, ),
        ),
        legend_opts=opts.LegendOpts(
            pos_top=25,
            pos_left='center',
            textstyle_opts=opts.TextStyleOpts(
                font_size=10,  # Change the font size of the legend
            ),
        )
    )

    # add data
    pie2.add("",
             [list(z) for z in zip(labels, counts_distribution)],
             label_opts=opts.LabelOpts(
                 position="inside",
                 formatter="{b|{b}: }{c}  {per|{d}%}",
                 border_width=0,
                 border_radius=0,
                 rich={
                     "b": {"fontSize": 15, "lineHeight": 20},
                     "per": {
                         "color": "#eee",
                         "padding": [1, 1],
                         "borderRadius": 0,
                     },
                 },
             ),
             )
    pie2_color = [speaker_color_map[label] for label in labels]
    pie2.set_colors(pie2_color)
    pie2.set_global_opts(
        title_opts=opts.TitleOpts(
            title="Segment Distribution",
            pos_top="0%",
            pos_left='center',
            title_textstyle_opts=opts.TextStyleOpts(font_size=15, ),
        ),
        legend_opts=opts.LegendOpts(
            pos_top=25,
            pos_left='center',
            textstyle_opts=opts.TextStyleOpts(
                font_size=10,  # Change the font size of the legend
            ),
        )
    )

    page.add(stacked_bar)
    page.add(pie1)
    page.add(pie2)
    bucket_name = f"session_{os.path.basename(json_file_path).split('_')[1]}"
    page.render(os.path.join(save_dir, f'{bucket_name}_speaker_diarization.html'))
    print(f"Speaker diarization visualization saved to '{bucket_name}_speaker_diarization.html'")


def plot_speaking_interaction_network(json_file_path, save_dir):
    """Creates and visualizes a network graph of speaking interactions using NetworkX and matplotlib."""
    speaker_interactions = calculate_speaker_interactions(json_file_path)
    session_name = os.path.basename(json_file_path).split('_')[1]

    # Create a directed graph
    G = nx.DiGraph()
    for src, targets in speaker_interactions.items():
        for tgt, freq in targets.items():
            G.add_edge(src, tgt, weight=freq)

    fixed_positions = None
    plt.figure(figsize=(10, 10))
    pos = nx.spring_layout(G, seed=42) if fixed_positions is None else fixed_positions

    # Calculate accumulative weights for each node
    accumulative_weights = {node: sum(weight for _, _, weight in G.edges(node, data='weight')) for node in G.nodes}

    # Node size parameters
    max_node_size = 2000
    min_node_size = 500

    # Normalize node sizes
    node_sizes = {}
    for node, weight in accumulative_weights.items():
        node_sizes[node] = min_node_size + (weight - 0) / (0.5 - 0) * (
                max_node_size - min_node_size) if weight > 0 else min_node_size

    # Draw nodes
    nx.draw_networkx_nodes(G, pos, node_size=[node_sizes[node] for node in G.nodes], alpha=0.2)

    # Draw edges
    # Identify curved edges
    curved_edges = [edge for edge in G.edges() if G.has_edge(edge[1], edge[0])]
    straight_edges = list(set(G.edges()) - set(curved_edges))

    # Draw straight edges
    for (u, v) in straight_edges:
        wt = G[u][v]['weight']
        nx.draw_networkx_edges(G, pos, edgelist=[(u, v)], width=weight_to_width(wt),
                               connectionstyle=f'arc3, rad = 0.2',
                               edge_color='moccasin', style='solid')

    # Draw curved edges with arc style
    for (u, v) in curved_edges:
        wt = G[u][v]['weight']
        nx.draw_networkx_edges(G, pos, edgelist=[(u, v)], width=weight_to_width(wt),
                               connectionstyle=f'arc3, rad = 0.2',
                               edge_color='orange', style='dashed')

    # Draw edge labels using the custom function
    ax = plt.gca()
    arc_rad = 0.25
    edge_weights = nx.get_edge_attributes(G, 'weight')
    curved_edge_labels = {edge: round(edge_weights[edge], 4) for edge in curved_edges}
    straight_edge_labels = {edge: round(edge_weights[edge], 4) for edge in straight_edges}
    my_draw_networkx_edge_labels(G, pos, ax=ax, edge_labels=curved_edge_labels, rotate=True, rad=arc_rad)
    my_draw_networkx_edge_labels(G, pos, ax=ax, edge_labels=straight_edge_labels, rotate=True, rad=arc_rad)

    # Draw node labels
    nx.draw_networkx_labels(G, pos, font_size=7)

    plt.title(f'Normalized Speaking Interaction Network ({session_name})')
    bucket_name = f"session_{os.path.basename(json_file_path).split('_')[1]}"
    plt.savefig(os.path.join(save_dir, f"{bucket_name}_nsin.png"), dpi=300)
    print(f"Normalized speaking interaction network saved to '{bucket_name}_nsin.png'")


def calculate_speaker_interactions(json_file_path):
    """Calculates the frequency of interactions between speakers in a conversation."""
    # Load JSON data from the provided file path
    with open(json_file_path, 'r') as file:
        data = json.load(file)
    normalize_factor = len(data)  # Normalized by number of segments

    # Initialize a dictionary to hold the interaction counts
    interaction_counts = {}

    # Function to update the interaction counts
    def update_interactions(cur_speakers, nex_speakers):
        for next_speaker in nex_speakers:
            if next_speaker not in interaction_counts:
                interaction_counts[next_speaker] = {}
            for current in cur_speakers:
                if current != next_speaker:
                    interaction_counts[next_speaker][current] = (interaction_counts[next_speaker].get(current, 0) + 1 /
                                                                 normalize_factor)

    # Iterate through the JSON data
    for i in range(len(data) - 1):
        current_segment = data[i]
        next_segment = data[i + 1]

        # Parse speakers from the current and next segments
        current_speakers = json.loads(current_segment['speakers'])
        next_speakers = json.loads(next_segment['speakers'])

        # Filter out 'silent' and 'unknown'
        current_speakers = [s for s in current_speakers if s not in ['silent', 'unknown']]
        next_speakers = [s for s in next_speakers if s not in ['silent', 'unknown']]

        # Update interaction counts
        update_interactions(current_speakers, next_speakers)

    return interaction_counts


def across_sessions_analysis_audio(logs_dir, visualizations_dir):
    """Analyzes and visualizes various conversation metrics across multiple sessions using saved log data."""
    session_data = []
    for bucket_dir_name in sorted(os.listdir(logs_dir)):
        bucket_dir_path = os.path.join(logs_dir, bucket_dir_name)
        session_name = bucket_dir_name.split('_')[1]
        if os.path.isdir(bucket_dir_path):
            apd, peq, nttc, sr = None, None, None, None
            for filename in os.listdir(bucket_dir_path):
                if filename.endswith("_recognition.json"):
                    json_path = os.path.join(bucket_dir_path, filename)
                    apd, peq, nttc, sr = analyze_recognition_log(json_path)
            session_data.append((session_name, apd, peq, nttc, sr))

    # Sorting the session data by session names
    session_data.sort(key=lambda x: x[0])

    # Unpacking the sorted data into separate lists
    session_names, apd_lst, peq_lst, nttc_lst, sr_lst = zip(*session_data)

    plot_across_sessions_analysis_audio(session_names, apd_lst, nttc_lst, sr_lst, peq_lst, visualizations_dir)
    plot_across_sessions_analysis_audio_interactive(session_names, apd_lst, nttc_lst, sr_lst, peq_lst,
                                                    visualizations_dir)


def analyze_recognition_log(json_file_path, unknown=False):
    """An updated version of analyze_conversation_metrics, offering refined analysis of conversation metrics."""
    # Read JSON file
    with open(json_file_path, 'r') as f:
        logs = json.load(f)

    if not unknown:
        # Filter out logs with 'unknown' speaker
        logs = [log for log in logs if 'unknown' not in log['speakers']]

    # Initialize variables
    turn_taking_count = 0
    current_silence_duration = 0
    silence_durations = []
    speaker_counter = Counter()
    speaker_durations = Counter()
    normalize_factor = len(logs)  # Normalized by number of segments

    # Loop through each log entry to calculate metrics
    for j in range(len(logs) - 1):
        log = logs[j]
        next_log = logs[j + 1]
        speakers = json.loads(log['speakers'])
        durations = json.loads(log['durations'])
        next_speakers = json.loads(next_log['speakers'])

        for i, speaker in enumerate(speakers):
            duration = durations[i]
            speaker_counter[speaker] += 1
            speaker_durations[speaker] += duration

            if speaker == "silent":
                current_silence_duration += duration
                # Reset and store silence duration if the next speaker is not silent
                if speaker not in next_speakers:
                    silence_durations.append(current_silence_duration)
                    current_silence_duration = 0
            # Turn-taking count, speaker A -> speaker B counts once, speaker A -> silent doesn't count
            if speaker not in next_speakers and 'silent' not in next_speakers:
                turn_taking_count += 1

        if j == len(logs) - 2:
            next_durations = json.loads(next_log['durations'])
            for i, speaker in enumerate(next_speakers):
                duration = next_durations[i]
                speaker_counter[speaker] += 1
                speaker_durations[speaker] += duration

                if speaker == 'silent':
                    current_silence_duration += duration
                    silence_durations.append(current_silence_duration)
                    current_silence_duration = 0

    # Calculate Average Pause Duration (APD)
    apd = sum(silence_durations) / len(silence_durations) if silence_durations else 0

    # Calculate Participation Equality using Gini Coefficient (PEQ)
    filtered_speaker_counter = {k: v for k, v in speaker_counter.items() if k not in ["silent", "unknown"]}
    n = len(filtered_speaker_counter)
    total_utterances = sum(filtered_speaker_counter.values())
    mean_utterances = total_utterances / n if n else 0

    if n == 0 or total_utterances == 0:  # Handle case where no known speakers are present
        peq = None
    else:
        gini_coefficient = sum(
            sum(abs(x - y) for y in filtered_speaker_counter.values())
            for x in filtered_speaker_counter.values()) / (2 * n * n * mean_utterances)
        peq = 1 - gini_coefficient

    # Normalized Turn Taking Count (NTTC)
    nttc = turn_taking_count / normalize_factor

    # Silence Ratio (SR)
    sr = speaker_durations["silent"] / sum(speaker_durations.values()) if "silent" in speaker_durations else 0

    # Return calculated metrics
    return apd, peq, nttc, sr


def plot_across_sessions_analysis_audio(session_names, apd_lst, nttc_lst, sr_lst, peq_lst, save_dir):
    """Creates and saves matplotlib plots of various audio analysis metrics across different sessions."""
    # Create subplots
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 12))

    # Plotting the first graph
    ax1.set_title("APD and NTTC")
    ax1.set_xlabel("Session")
    ax1.set_ylabel("Average Pause Duration (seconds/silence pause)", color='tab:blue')
    ax1.plot(session_names, apd_lst, 'o-', color='tab:blue', label='APD')
    ax1.tick_params(axis="y", labelcolor="tab:blue")

    ax1_twin = ax1.twinx()
    ax1_twin.set_ylabel("Normalized Turn Taking Count (turns/segment)", color='tab:red')
    ax1_twin.plot(session_names, nttc_lst, 'x-', color='tab:red', label='NTTC')
    ax1_twin.tick_params(axis="y", labelcolor="tab:red")

    # Plotting the second graph
    ax2.set_title("SR and PEQ")
    ax2.set_xlabel("Session")
    ax2.set_ylabel("Silence Ratio (%)", color='tab:green')
    ax2.plot(session_names, sr_lst, 's-', color='tab:green', label='SR')
    ax2.tick_params(axis="y", labelcolor="tab:green")

    ax2_twin = ax2.twinx()
    ax2_twin.set_ylabel("Participation Equality (%)", color='tab:purple')
    ax2_twin.plot(session_names, peq_lst, 'd-', color='tab:purple', label='PEQ')
    ax2_twin.tick_params(axis="y", labelcolor="tab:purple")

    # Show legends
    ax1.legend(loc='upper left')
    ax1_twin.legend(loc='upper right')
    ax2.legend(loc='upper left')
    ax2_twin.legend(loc='upper right')

    # Show plot
    fig.suptitle("Audio Analysis Metrics Across Sessions")
    fig.tight_layout()
    plt.savefig(os.path.join(save_dir, 'audio_analysis_across_sessions.png'), dpi=300)
    print("Audio analysis metrics across sessions saved to 'audio_analysis_across_sessions.png'")


def plot_across_sessions_analysis_audio_interactive(session_names, apd_lst, nttc_lst, sr_lst, peq_lst, save_dir):
    """Visualize audio analysis metrics across sessions with pyecharts."""
    # Generate visualization page
    page = Page(page_title="audio metrics analysis across sessions", layout=Page.SimplePageLayout)

    def format_list(data):
        return [f"{x:.4f}" if x is not None else '0' for x in data]

    apd_lst = format_list(apd_lst)
    nttc_lst = format_list(nttc_lst)
    sr_lst = format_list(sr_lst)
    peq_lst = format_list(peq_lst)
    yaxis_min_js = JsCode("function(value){return Math.round(value.min * 0.8 * 1000) / 1000;}")
    yaxis_max_js = JsCode("function(value){return Math.round(value.max * 1.2 * 1000) / 1000;}")

    # Create Line chart for APD and NTTC
    line1 = (
        Line(init_opts=opts.InitOpts(width="1500px", height="400px"))
        .add_xaxis(session_names)
        .add_yaxis("APD", apd_lst, is_smooth=False, yaxis_index=0)
        .add_yaxis("NTTC", nttc_lst, is_smooth=False, yaxis_index=1)
        .extend_axis(
            yaxis=opts.AxisOpts(
                name="Normalized Turn Taking Count (turns/segment)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
                type_="value",
                position="right",
            )
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title="APD and NTTC", pos_top="0%", pos_left="center"),
            legend_opts=opts.LegendOpts(
                pos_top="7%",
                item_width=20,
                item_height=10,
                textstyle_opts=opts.TextStyleOpts(font_size=10)
            ),
            yaxis_opts=opts.AxisOpts(
                name="Average Pause Duration (seconds/silent pause)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
            ),
            xaxis_opts=opts.AxisOpts(name="Session", name_location="middle", name_gap=30)
        )
    )

    # Create Line chart for PEQ and SR
    line2 = (
        Line(init_opts=opts.InitOpts(width="1500px", height="400px"))
        .add_xaxis(session_names)
        .add_yaxis("SR", sr_lst, is_smooth=False, yaxis_index=0)
        .add_yaxis("PEQ", peq_lst, is_smooth=False, yaxis_index=1)
        .extend_axis(
            yaxis=opts.AxisOpts(
                name="Participation Equality (%)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
                type_="value",
                position="right",
            )
        )
        .set_global_opts(
            title_opts=opts.TitleOpts(title="SR and PEQ", pos_top="0%", pos_left="center"),
            legend_opts=opts.LegendOpts(
                pos_top="7%",
                item_width=20,
                item_height=10,
                textstyle_opts=opts.TextStyleOpts(font_size=10)
            ),
            yaxis_opts=opts.AxisOpts(
                name="Silence Ratio (%)",
                min_=yaxis_min_js,
                max_=yaxis_max_js,
            ),
            xaxis_opts=opts.AxisOpts(name="Session", name_location="middle", name_gap=30)
        )
    )

    page.add(line1)
    page.add(line2)
    page.render(os.path.join(save_dir, f'audio_analysis_across_sessions.html'))
    print("Audio analysis metrics across sessions saved to 'audio_analysis_across_sessions.html'")



import os

import numpy as np
from scipy.io.wavfile import read, write


def read_file(filename):
    """Read wave file as mono.

    Args:
        filename (str) : wave file / path.

    Returns:
        tuple of sampling rate and audio data.
    """
    fr, sig = read(filename=filename)
    if sig.ndim == 1:
        samples = sig
    else:
        samples = sig[:, 0]
    if sig.dtype == np.int16:
        # print("This is 16-bit PCM audio. Normalize by dividing by 32767.0.")
        samples = samples / 32767.0
    return fr, samples


def write_file(output_file_path, input_file_name, name_attribute, sig, fs):
    """Write wave file as mono.

    Args:
        output_file_path (str) : path to save resulting wave file to.
        input_file_name  (str) : name of processed wave file,
        name_attribute   (str) : attribute to add to output file name.
        sig            (array) : signal/audio array.
        fs               (int) : sampling rate.

    Outputs:
        Save wave file to output_file_path.
    """
    # Convert back to int16 if necessary (will decrease accuracy)
    if sig.dtype == np.float64:
        sig = np.int16(sig * 32767)

    # Set up the output file name
    file_name = os.path.basename(input_file_name).split(".wav")[0] + name_attribute
    file_path = os.path.join(output_file_path, file_name)
    write(filename=file_path, rate=fs, data=sig)
    # print("Writing data to " + file_path + ".")



import base64
import configparser
import gc
import os

import torch
from flask import request, jsonify
from modelscope.pipelines import pipeline
from modelscope.utils.constant import Tasks

from openmmla_audio.utils.audio_utils import write_frames_to_file
from openmmla_audio.utils.logger_utils import get_logger


class SpeechSeparator:
    """Speech separator separates the audio signals from a mixed signal. It receives audio signal from base station,
    and sends back two separated signals. It can help to differentiate a speaker's voice from an overlapped speaking
    environment."""

    def __init__(self, project_dir, config_path=None, use_cuda=True):
        """Initialize the speech separator.

        Args:
            project_dir: the project directory.
            use_cuda: whether to use CUDA or not.
        """
        self.server_logger_dir = os.path.join(project_dir, 'server/logger')
        self.server_file_folder = os.path.join(project_dir, 'server/temp')
        os.makedirs(self.server_logger_dir, exist_ok=True)
        os.makedirs(self.server_file_folder, exist_ok=True)

        self.logger = get_logger('modelscope', os.path.join(self.server_logger_dir, 'separate_server.log'))
        self.cuda_enable = use_cuda and torch.cuda.is_available()

        if config_path:
            if os.path.isabs(config_path):
                self.config_path = config_path
            else:
                self.config_path = os.path.join(project_dir, config_path)
        else:
            self.config_path = os.path.join(project_dir, 'conf/audio_base.ini')

        # Check if the configuration file exists
        if not os.path.exists(self.config_path):
            raise FileNotFoundError(f"Configuration file not found at {self.config_path}")

        config = configparser.ConfigParser()
        config.read(self.config_path)
        sp_model = config['Server']['sp_model']
        sp_model_local = config['Server']['sp_model_local']
        self.device = 'gpu' if self.cuda_enable else 'cpu'

        # Initialize speech separation model
        try:
            self.separation_model = pipeline(Tasks.speech_separation, device=self.device, model=sp_model)
        except ValueError:
            self.separation_model = pipeline(Tasks.speech_separation, device=self.device, model=sp_model_local)

    def separate_speech(self):
        """Separate the speech from the audio.

        Returns:
            A tuple containing the JSON response and status code.
        """
        if request.files:
            try:
                base_id = request.values.get('base_id')
                audio_file = request.files['audio']
                audio_file_path = os.path.join(self.server_file_folder, f'separate_audio_{base_id}.wav')
                write_frames_to_file(audio_file_path, audio_file.read(), 1, 2, 8000)

                result = self.separation_model(audio_file_path)
                processed_bytes_streams = []

                for signal in result['output_pcm_list']:
                    encoded_bytes_stream = base64.b64encode(signal).decode('utf-8')
                    processed_bytes_streams.append(encoded_bytes_stream)

                # Return the processed audio data
                return jsonify({"processed_bytes_streams": processed_bytes_streams}), 200
            except Exception as e:
                self.logger.error(f"during speech separation, {e} happens.")
                return jsonify({"error": str(e)}), 500
            finally:
                torch.cuda.empty_cache()
                gc.collect()
        else:
            return jsonify({"error": "No audio file provided"}), 400



import os
import signal
import subprocess


def find_process_using_port(port):
    """Find the PID of the process using the specified port."""
    command = f"lsof -i tcp:{port} | awk 'NR!=1 {{print $2}}'"
    result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    pids = result.stdout.strip().split('\n')
    unique_pids = list(set(pid for pid in pids if pid.isdigit()))
    return unique_pids


def kill_process(pid):
    """Kill the process with the given PID."""
    try:
        os.kill(int(pid), signal.SIGKILL)
        print(f"Process {pid} killed successfully.")
    except OSError as error:
        print(f"Error: {error}")


def free_port(port):
    pids = find_process_using_port(port)
    if pids:
        print(f"Processes using port {port}: {pids}")
        for pid in pids:
            kill_process(pid)
    else:
        print(f"No process is using port {port}.")
